# FILE: connection.tcl
#
# An object which manages asynchronous communication over a Net_Link.
#
# Query, expects reply
# Tell, just sends message
#
# The Tcl event loop must be running for this event-driven object to function.
#

Oc_Class Net_Connection {

    # The link which manages lower level of communications.
    private variable link

    # The socket we hand off to our link
    public variable socket 

    # The hostname we hand off to our link
    public variable hostname 

    # The port we hand off to our link
    public variable port 
    
    # The protocol used to handle queries received over this connection.
    public variable protocol
    private variable myProtocol

    # State of query handling protocol state machine -- move this into protocol
    private variable state

    # Log for logging messages
    public variable log {
        if {[catch {$log Append "Logging messages generated by $this."}]} {
            error "$log is not able to log messages."
        }
        if {[info exists logHandler]} {
            $logHandler delete
        }
        set logHandler [em_eventHandler new $log delete "$this Delete"]
        if {$imadelog} {
            $oldlog Delete
        }
        set oldlog $log
    }
    private variable oldlog = Oc_Nop
    private variable logHandler
    private variable imadelog = 0

    # Received message
    private variable message

    # query id
    private variable id = 0

    # Ready to send/receive?
    private variable ready = 0

    # Array of timeout events
    private array variable toe = {}
    public variable timeout = 60000

    # My binding tags
    private variable bindtags
    private variable myEventHandlers = {}

    # Constructor is a non-blocking operation.  
    Constructor { args } {
        eval $this Configure $args
        if {![info exists log]} {
            # Provide a default log
            $this Configure -log [em_log New _ -title "Connection $this Log"]
            set imadelog 1
        }
        # Enable event generation
        set bindtags [list $this $class]
        if {[info exists socket]} {
            # Assume a server
            if {![info exists protocol]} {
                error "Server connections need a query-handling protocol"
            }
            set myProtocol [$protocol Clone $this]
            lappend myEventHandlers [em_eventHandler new $myProtocol delete \
                    [list $this Delete]]
            Net_Link New link -socket $socket -log $log
            lappend myEventHandlers [em_eventHandler new $link Ready \
                    [list $this ServerLinkReady]]
        } elseif {[info exists hostname] && [info exists port]} {
            # Assume a client
            Net_Link New link -hostname $hostname -port $port -log $log
            lappend myEventHandlers [em_eventHandler new $link Ready \
                    [list $this LinkReady]]
        } else {
            error "Missing required options: -hostname and -port or -socket"
        }
        lappend myEventHandlers [em_eventHandler new $link delete \
                [list $this Delete]]
    }

    method ServerLinkReady {} {
        $log Append "Put [concat tell [$myProtocol Cget -name]]"
        $link Put [concat tell [$myProtocol Cget -name]]
        $this LinkReady
    }

    method LinkReady {} {
        lappend myEventHandlers [em_eventHandler new $link Readable \
                [list $this Receive]]
        set hostname [$link Cget -hostname]
        set port [$link Cget -port]
        set state start
        set ready 1
        em_eventHandler Generate Ready $bindtags
    }

    # Every line of data received over the link gets handled by this
    # method.  It has to parse the message and direct it to its 
    # destination.
    method Receive {} {
        # NOTE: The function of the Net_Link is to transmit and receive
        # strings.  The Net_Connection wants to interpret these strings
        # as Tcl lists.  Not all string are a valid Tcl list, so we have
        # to be sure the line of input we receive is tested, or somehow
        # interpreted as a valid Tcl list.  I would like to use the
        # split and join commands to translate back and forth between
        # lists and strings but I can't because contrary to their 
        # documentation, they are *not* inverses of each other.
        # For example:
        # % split [join {1 {2 3} {4 5 6}}]
        # 1 2 3 4 5 6
        # The list structure can be completely changed. Grrrrrrr..
        # It seems the best we can do is use the Tcl list commands to
        # at least test the the string we receive has a valid interpretation
        # as a Tcl list, and log an error and bail if it does not.

        # Get the string
        set rawmessage [$link Get]
        $log Append "Received raw message: $rawmessage"
        # Test for list structure 
        # The 'eval' is required to expand the string $rawmessage into
        # whatever words (list elements) it contains.
        if {[catch {eval list $rawmessage} msg]} {
            Oc_Log Log "Message not a Tcl list: $rawmessage" \
                    warning Net_Connection
            return
        }
        if {![llength $msg]} {
            # Ignore empty messages
            return
        }
        switch -- [lindex $msg 0] {
            reply {
                if {[llength $msg] == 1} {
                    Oc_Log Log "Received bad message: $msg" \
                            warning Net_Connection
                    return
                }
                set rid [lindex $msg 1]
                if {![catch {set toe($rid)} ev]} {
                    after cancel $ev
                    unset toe($rid)
                }
                set message [lrange $msg 2 end]
                em_eventHandler Generate Reply$rid $bindtags
            }
            query {
                if {[llength $msg] == 1} {
                    Oc_Log Log "Received bad message: $msg" \
                            warning Net_Connection
                    return
                }
                set id [lindex $msg 1]
                set message [lrange $msg 2 end]
                $log Append "Handling query: $message"
                # This doesn't allow for event-driven query handling.
                # Should it?  If so, should it timeout?
		set reply {}   ;# Safety
		# Use local variable 'localstate' to work around possible
		# bug in bytecode compiler in Tcl 8.0 - 8.0p2
		# ... or perhaps a more mundane re-entrancy problem
                foreach {localstate reply} \
                        [$myProtocol Reply $state $message] {}
		# Must check existence in case the call above
		# called our destructor.
		if {[info exists state]} {
                    set state $localstate
                    $log Append "Put [concat reply $id $reply]"
                    $link Put [concat reply $id $reply]
                    if {[string match close $state]} {
                        $log Append "Closing connection on remote command"
                        $this Delete
                    }
		}
            }
            tell {
                set message [lrange $msg 1 end]
                em_eventHandler Generate Readable $bindtags
            }
            default {
                # Log non-conforming messages, but continue.
                Oc_Log Log "Received bad message: $msg" \
                        warning Net_Connection
            }
        }
    }

    method Get {} {
        return $message
    }

    method Tell { args } {
        if {!$ready} {
            error "connection not ready"
        }
        $log Append "Put [concat tell $args]"
        $link Put [concat tell $args]
    }

    method Query { args } {
        if {!$ready} {
            error "connection not ready"
        }
        incr id
        array set toe [list $id [after $timeout $this Timeout $id]]
        $log Append "Put [concat query $id $args]"
        $link Put [concat query $id $args]
        return $id
    }

    method Timeout { qid } {
        $log Append "Timeout waiting for reply to query $qid"
        em_eventHandler Generate Timeout$qid $bindtags
    }

    Destructor {
        if {[info exists bindtags]} {
            em_eventHandler Generate delete $bindtags
        }
        # delete handlers before deleting event-generating members
        foreach i [array names toe] {
            after cancel $toe($i)
            unset toe($i)
        }
        if {[info exists myEventHandlers]} {
            foreach handler $myEventHandlers {
                $handler delete
            }
        }
        if {[info exists myProtocol]} {
            $myProtocol Delete
        }
        if {[info exists link]} {
            $link Delete
        }
        if {[info exists log]} {
            $log Append "Deleting $this ..."
            $this Configure -log Oc_Nop
        }
    }

}
